#include "../include/app.h"
#include <GL/gl.h>

void init_app(App *app, uint w, uint h)
{
    int err;

    err = SDL_Init(SDL_INIT_EVERYTHING);
    if (err != 0) {
        printf("Failed to initialize SDL. Error: %s\n",
               SDL_GetError());
        exit(1);
    }

    app->window = SDL_CreateWindow(
                    "test",
                    SDL_WINDOWPOS_CENTERED, 
                    SDL_WINDOWPOS_CENTERED,
                    w, h, SDL_WINDOW_OPENGL);

    if (app->window == NULL) {
        printf("Failed to create window.\n");
        exit(1);
    }

    app->context = SDL_GL_CreateContext(app->window);
    if (app->context == NULL) {
        printf("Failed to create OpenGL context.\n");
        exit(1);
    }
    app->is_running = true;

    //SDL_ShowCursor(SDL_DISABLE);
    init_camera(&app->camera);
    init_opengl();
    reshape(w, h);

    init_camera(&(app->camera));

    app->surface = IMG_Load("assets/block.png"); 
    app->tid = 0;
    
    if(app->surface==NULL) {
            printf("error\n");
            exit(1);
    }
    glGenTextures(1, &app->tid);
    glBindTexture(GL_TEXTURE_2D, app->tid);
        
    int Mode = GL_RGB;

    if(app->surface->format->BytesPerPixel == 4) {
        Mode = GL_RGBA;
    }
        
    glTexImage2D(GL_TEXTURE_2D, 0, Mode, app->surface->w, app->surface->h, 0, Mode, GL_UNSIGNED_BYTE, app->surface->pixels);
        
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    glBindTexture(GL_TEXTURE_2D, app->tid);
}

void update_app(App* app)
{
    app->camera.speed.z -= GRAVITY * (app->frame_time/1000);

    if(app_check_collision(app))
        camera_set_vertical_speed(&app->camera, 0.0);
    update_camera(&(app->camera), app->frame_time/1000);
}

void init_opengl()
{
    // https://gitlab.com/imre-piller/me-courses/ alapjan.
    glShadeModel(GL_SMOOTH);
 
    glEnable(GL_NORMALIZE);
    glEnable(GL_AUTO_NORMAL);
    
    glClearColor(0.1, 0.1, 0.1, 1.0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glEnable(GL_DEPTH_TEST);
    glClearDepth(1.0);
    glEnable(GL_TEXTURE_2D);
}

void reshape(GLsizei width, GLsizei height)
{
    int x, y, w, h;
    double ratio;

    ratio = (double)width / height;
    if (ratio > VIEWPORT_RATIO) {
        w = (int)((double)height * VIEWPORT_RATIO);
        h = height;
        x = (width - w) / 2;
        y = 0;
    }
    else {
        w = width;
        h = (int)((double)width / VIEWPORT_RATIO);
        x = 0;
        y = (height - h) / 2;
    }

    glViewport(x, y, w, h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(
        -.16, .16,
        -.09, .09,
        .1, 1000
    );
}

void handle_events(App* app)
{
    SDL_Event event;
    static bool is_mouse_down = false;
    static int mouse_x = 0;
    static int mouse_y = 0;
    int x;
    int y;

    while (SDL_PollEvent(&event)) {
        switch (event.type) {
        case SDL_KEYDOWN:
            switch (event.key.keysym.scancode) {
            case SDL_SCANCODE_ESCAPE:
                app->is_running = false;
                break;
            case SDL_SCANCODE_W:
                set_camera_speed(&(app->camera), SPEED);
                break;
            case SDL_SCANCODE_S:
                set_camera_speed(&(app->camera), -SPEED);
                break;
            case SDL_SCANCODE_A:
                set_camera_side_speed(&(app->camera), SPEED);
                break;
            case SDL_SCANCODE_D:
                set_camera_side_speed(&(app->camera), -SPEED);
                break;
            case SDL_SCANCODE_SPACE:
                if (app_check_collision(app))
                    app->camera.position.z+=0.3;
                camera_set_vertical_speed(&app->camera, SPEED);
                break;
            case SDL_SCANCODE_LSHIFT:
                camera_set_vertical_speed(&app->camera, -SPEED);
                break;
            default:
                break;
            }
            break;
        case SDL_KEYUP:
            switch (event.key.keysym.scancode) {
            case SDL_SCANCODE_W:
            case SDL_SCANCODE_S:
                set_camera_speed(&(app->camera), 0);
                break;
            case SDL_SCANCODE_A:
            case SDL_SCANCODE_D:
                set_camera_side_speed(&(app->camera), 0);
                break;
            case SDL_SCANCODE_SPACE:
            case SDL_SCANCODE_LSHIFT:
                camera_set_vertical_speed(&app->camera, -0);
            default:
                break;
            }
            break;
        case SDL_MOUSEBUTTONDOWN:
            app_break_block(app);
            break;
        case SDL_MOUSEMOTION:
            SDL_GetMouseState(&x, &y);
            rotate_camera(&(app->camera), mouse_x - x, mouse_y - y);
            mouse_x = x;
            mouse_y = y;
            break;
        case SDL_MOUSEBUTTONUP:
            is_mouse_down = false;
            break;
        case SDL_QUIT:
            app->is_running = false;
            break;
        default:
            break;
        }
    }

}

void draw_cube(App *app, float x, float y, float z)
{
    /*x*=2;
    y*=2;
    z*=2;
    */
    glBegin( GL_QUADS );

        glColor3f(1, 1, 1); 
        glTexCoord2f(0, 0); glVertex3f(x+1.0, y, z-1.0 );  
        glTexCoord2f(0, 1); glVertex3f(x, y, z-1.0);
        glTexCoord2f(1, 1); glVertex3f(x, y, z); 
        glTexCoord2f(1, 0); glVertex3f(x+1.0, y, z);

        glColor3f(1, 1, 1);
        glTexCoord2f(0, 0); glVertex3f(x+1.0, y-1.0, z-1.0);
        glTexCoord2f(0, 1); glVertex3f(x, y-1.0, z-1.0);
        glTexCoord2f(1, 1); glVertex3f(x, y-1.0, z);
        glTexCoord2f(1, 0); glVertex3f(x+1.0, y-1.0, z); 

        glColor3f(1, 1, 1);
        glTexCoord2f(0, 0); glVertex3f(x+1.0, y, z);
        glTexCoord2f(0, 1); glVertex3f(x, y, z);
        glTexCoord2f(1, 1); glVertex3f(x, y-1.0, z);
        glTexCoord2f(1, 0); glVertex3f(x+1.0, y-1.0, z);

        glColor3f(1, 1, 1);
        glTexCoord2f(0, 0); glVertex3f(x, y, z-1.0); // Top-Right of back face
        glTexCoord2f(0, 1); glVertex3f(x+1.0, y, z-1.0); // Top-Left of back face
        glTexCoord2f(1, 1); glVertex3f(x+1.0, y-1.0, z-1.0); // Bottom-Left of back face
        glTexCoord2f(1, 0); glVertex3f(x, y-1.0, z-1.0); // Bottom-Right of back face

        glColor3f(1, 1, 1);
        glTexCoord2f(0, 0); glVertex3f(x, y, z);  // Top-Right of left face
        glTexCoord2f(0, 1); glVertex3f(x, y, z-1.0);  // Top-Left of left face
        glTexCoord2f(1, 1); glVertex3f(x, y-1.0, z-1.0);  // Bottom-Left of left face
        glTexCoord2f(1, 0); glVertex3f(x, y-1.0, z);  // Bottom-Right of left face

        glColor3f(1, 1, 1);
        glTexCoord2f(0, 0); glVertex3f(x+1.0, y, z);  // Top-Right of left face
        glTexCoord2f(0, 1); glVertex3f(x+1.0, y, z-1.0);  // Top-Left of left face
        glTexCoord2f(1, 1); glVertex3f(x+1.0, y-1.0, z-1.0);  // Bottom-Left of left face
        glTexCoord2f(1, 0); glVertex3f(x+1.0, y-1.0, z);  // Bottom-Right of left face
    glEnd();
    glFlush();
}  

float calc_frame_time(struct timespec *start, struct timespec *end)
{
    float elapsed_time = (end->tv_sec - start->tv_sec)*1000;
    
    elapsed_time += (float)(end->tv_nsec - start->tv_nsec)/1000000;

    return elapsed_time;
}

void app_generate_world(App *app, int a)
{
    app->chunk_count = a*a;
    app->chunks = malloc(a*a*sizeof(Chunk));
    memset(app->chunks, 0, a*a*sizeof(Chunk));

    unsigned char ptable[512];
    noise_init_ptable(ptable);

    for (int i = 0; i < a; i++) {
        for (int j = 0; j < a; j++) {
            Chunk c;
            c.start_pos.x = i*CHUNK_MAX_X;
            c.start_pos.y = j*CHUNK_MAX_Y;
            c.start_pos.z = 0;
            for (int x = 0; x < CHUNK_MAX_X; x++) {
                for (int y = 0; y < CHUNK_MAX_Y; y++) {
                    //printf("x: %d y:%d noise: %f\n", x, y, round((noise_noise2(ptable,
                    //      (c.start_pos.x+x)*0.03, (c.start_pos.y+y)*0.03) + 1) * 0.5 * 15);
                    int mz = round((noise_noise2(ptable, (c.start_pos.x+x)*0.03, (c.start_pos.y+y)*0.03) + 1) * 0.5 * 15);
                    if (mz < 0)
                        mz = 0;
                    else if (mz > CHUNK_MAX_Z)
                        mz = CHUNK_MAX_Z-1;
                    for (int z = 0; z < mz; z++)
                        c.blocks[x][y][z].type = BLOCKTYPE_STONE;
                    for (int z = mz; z < CHUNK_MAX_Z; z++)
                        c.blocks[x][y][z].type = BLOCKTYPE_AIR;
                }
            }
            chunk_set_blocks_visibility(&c);
            chunk_create_displayl(app, &c);

            app->chunks[i+(j*a)] = c;
        }
    }
}

int app_get_current_chunk_index(App *app)
{
    for (int i = 0; i < app->chunk_count; i++) {
        for (int x = 0; x < CHUNK_MAX_X; x++) {
            for (int y = 0; y < CHUNK_MAX_Y; y++) {
                for (int z = 0; z < CHUNK_MAX_Z; z++) {
                    if (app->camera.position.x < app->chunks[i].start_pos.x + CHUNK_MAX_X &&
                        app->camera.position.x > app->chunks[i].start_pos.x &&
                        app->camera.position.y < app->chunks[i].start_pos.y + CHUNK_MAX_Y &&
                        app->camera.position.y > app->chunks[i].start_pos.y) {
                            return i;
                    }
                }
            }
        }
    }

    return -1;
}

bool is_block(Block b)
{
    if(b.type != BLOCKTYPE_AIR)
        return true;
    return false;
}

bool app_check_collision(App *app)
{
    int i = app_get_current_chunk_index(app);

    if (i < 0)
        return false;

    vec3f c_pos = app->camera.position;
    vec3i c_index = { floor(c_pos.x), floor(c_pos.y), floor(c_pos.z) };

    for (int x = c_index.x-1; x < c_index.x+1; x++) {
        for (int y = c_index.y-1; y < c_index.y+1; y++) {
            for (int z = c_index.z-1; y < c_index.z+1; z++) {
                if(!check_index(x, y, z))
                    continue;
                if(is_block(app->chunks[i].blocks[x][y][z])) {
                     
                } 
            }
        }
    }

    return false;
}

int check_intersection_block(App *app, vec3i *retv, int *chunk_index)
{

    printf("%s\n", __func__);
    int i = app_get_current_chunk_index(app);

    *chunk_index = i;
    vec3f c_pos = app->camera.position;
    vec3f c_dir = get_camera_dir_vec3f(&app->camera);

    if (i < 0)
        return -1;

    for (int x = 0; x < CHUNK_MAX_X; x++) {
        for (int y = 0; y < CHUNK_MAX_Y; y++) {
            for (int z = 0; z < CHUNK_MAX_Z; z++) {
                if (app->chunks[i].blocks[x][y][z].visible) {
                    for (float j = 0; j < 8; j+=0.2) {
                        vec3f ray = {(c_pos.x - app->chunks[i].start_pos.x) + c_dir.x * j,
                                     (c_pos.y - app->chunks[i].start_pos.y) + c_dir.y * j,
                                     (c_pos.z - app->chunks[i].start_pos.z) + c_dir.z * j};
                        printf("ray x: %f y: %f z: %f\n", ray.x, ray.y, ray.z);
                        vec3i rayi = {round(ray.x/1), round(ray.y/1), round(ray.z/1)};

                        if (is_block(app->chunks[i].blocks[rayi.x][rayi.y][rayi.z])) {
                            retv->x = rayi.x;
                            retv->y = rayi.y;
                            retv->z = rayi.z;
                            return 0;
                        }
                    }
                } 
            }
        }
    }
    return -1;
}

void app_break_block(App *app)
{
    printf("%s\n", __func__);
    vec3i block;
    int i;
    int ret = check_intersection_block(app, &block, &i);

    printf("chunk index: %d\n", i);
    printf("player x: %f y: %f z: %f\n", app->camera.position.x, app->camera.position.y, app->camera.position.z);
    printf("chunk %d. min x: %f y: %f max x: %f y: %f\n", i, app->chunks[i].start_pos.x,
            app->chunks[i].start_pos.y, app->chunks[i].start_pos.x+CHUNK_MAX_X, app->chunks[i].start_pos.y+CHUNK_MAX_Y);
    printf("broken block x: %f y: %f z: %f\n", block.x+app->chunks[i].start_pos.x, block.y+app->chunks[i].start_pos.y, block.z+app->chunks[i].start_pos.z);
    if(ret < 0)
        return;

    app->chunks[i].blocks[block.x][block.y][block.z].type = BLOCKTYPE_AIR;
    chunk_update(app, &(app->chunks[i]));
}

