#include "../include/camera.h"

#include <GL/gl.h>
#include <SDL2/SDL.h>

#include <math.h>

void init_camera(Camera* camera)
{
    camera->position.x = 5.0;
    camera->position.y = 5.0;
    camera->position.z = 60.0;
    camera->rotation.x = 0.0;
    camera->rotation.y = 0.0;
    camera->rotation.z = 0.0;
    camera->speed.x = 0.0;
    camera->speed.y = 0.0;
    camera->speed.z = 0.0;
}

void update_camera(Camera* camera, double time)
{    
    double angle;
    double side_angle;

    angle = degree_to_radian(camera->rotation.z);
    side_angle = degree_to_radian(camera->rotation.z + 90.0);

    camera->position.x += cos(angle) * camera->speed.y * time;
    camera->position.y += sin(angle) * camera->speed.y * time;
    camera->position.z += camera->speed.z * time;
    
    camera->position.x += cos(side_angle) * camera->speed.x * time;
    camera->position.y += sin(side_angle) * camera->speed.x * time;
    camera->position.z += camera->speed.z * time;
}

void set_view(const Camera* camera)
{
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glRotatef(-(camera->rotation.x + 90), 1.0, 0, 0);
    glRotatef(-(camera->rotation.z - 90), 0, 0, 1.0);
    glTranslatef(-camera->position.x, -camera->position.y, -camera->position.z);
}
void set_view_fix(const Camera* camera)
{
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glRotatef(-20, 1.0, 0, 0);
    glRotatef(-30, 0, 0, 1.0);
    glTranslatef(-5, -5, -50);
}
void rotate_camera(Camera* camera, double horizontal, double vertical)
{
    camera->rotation.z += horizontal;
    camera->rotation.x += vertical;

    if (camera->rotation.z < 0) {
        camera->rotation.z += 360.0;
    }

    if (camera->rotation.z > 360.0) {
        camera->rotation.z -= 360.0;
    }

    if (camera->rotation.x < 0.0) {
        camera->rotation.x += 360.0;
    }

    if (camera->rotation.x > 360.0) {
        camera->rotation.x -= 360.0;
    }
}

void set_camera_speed(Camera* camera, double speed)
{
    camera->speed.y = speed;
}

void camera_set_vertical_speed(Camera* camera, double speed)
{
    camera->speed.z = speed;
}

void set_camera_side_speed(Camera* camera, double speed)
{
    camera->speed.x = speed;
}

vec3f get_camera_dir_vec3f(Camera *camera)
{
    vec3f dir = {cos(degree_to_radian(camera->rotation.x)) *
                 cos(degree_to_radian(camera->rotation.z)), 
                 sin(degree_to_radian(camera->rotation.z)) *
                 cos(degree_to_radian(camera->rotation.x)),
                 sin(degree_to_radian(camera->rotation.x))};

    vec3f_normalize(&dir);
    return dir;
}

